from hive.common import ArgumentParser

from .. import tell_user


class TellUserArgumentParser(ArgumentParser):
    def add_format_argument(self, format: str, default: str = "text"):
        help = f'send messages as format "{format.upper()}"'
        if format == default:
            help = f"{help} [the default if no format is specified]"
        self.add_argument(
            f"--{format}",
            action="store_const",
            dest="format", default=default, const=format,
            help=help,
        )


def main():
    parser = TellUserArgumentParser(
        description="Post messages to Hive's Matrix room.",
    )
    parser.add_argument(
        "message", metavar="MESSAGE",
        help="message to post"
    )
    for format in ("text", "html", "markdown", "code", "emojize"):
        parser.add_format_argument(format)
    args = parser.parse_args()

    tell_user(args.message, format=args.format)



from pika.exceptions import UnroutableError

from .channel import Channel, Notifier
from .connection import Connection
from .message_bus import MessageBus, Producer

DEFAULT_MESSAGE_BUS = MessageBus()

blocking_connection = DEFAULT_MESSAGE_BUS.blocking_connection
send_to_queue = DEFAULT_MESSAGE_BUS.send_to_queue
tell_user = DEFAULT_MESSAGE_BUS.tell_user

producer_connection = DEFAULT_MESSAGE_BUS.producer_connection



import os

from dataclasses import dataclass, field
from typing import Optional

from pika import (
    BlockingConnection,
    ConnectionParameters,
    PlainCredentials,
)
from pika.exceptions import AMQPConnectionError

from rstream import Producer

from hive.config import read as read_config

from .connection import Connection


@dataclass
class MessageBus:
    host: str = field(
        default_factory=lambda: os.environ.get(
            "RABBITMQ_HOST", "rabbit"),
    )
    port: int = field(
        default_factory=lambda: int(os.environ.get(
            "RABBITMQ_PORT",
            str(ConnectionParameters.DEFAULT_PORT))),
    )
    credentials_key: str = "rabbitmq"

    @property
    def credentials(self) -> PlainCredentials:
        env = read_config(self.credentials_key)
        return PlainCredentials(
            env["RABBITMQ_DEFAULT_USER"],
            env["RABBITMQ_DEFAULT_PASS"],
        )

    # Queues

    def queue_connection_parameters(
            self,
            *,
            host: Optional[str] = None,
            port: Optional[int] = None,
            credentials: Optional[PlainCredentials] = None,
            connection_attempts: int = 5,  # * (socket_timeout
            retry_delay: float = 2.0,  # ...    + retry_delay) = 60 seconds
            heartbeat: int = 600,
            blocked_connection_timeout: int = 300,
            **kwargs
    ) -> ConnectionParameters:
        if not host:
            host = self.host
        if not port:
            port = self.port
        if not credentials:
            credentials = self.credentials

        return ConnectionParameters(
            host=host,
            port=port,
            credentials=credentials,
            connection_attempts=connection_attempts,
            retry_delay=retry_delay,
            heartbeat=heartbeat,
            blocked_connection_timeout=blocked_connection_timeout,
            **kwargs
        )

    def blocking_connection(self, **kwargs) -> Connection:
        on_channel_open = kwargs.pop("on_channel_open", None)
        params = self.queue_connection_parameters(**kwargs)
        try:
            return Connection(
                BlockingConnection(params),
                on_channel_open=on_channel_open,
            )
        except AMQPConnectionError as e:
            e = getattr(e, "args", [None])[0]
            e = getattr(e, "exception", None)
            if isinstance(e, (ConnectionRefusedError, TimeoutError)):
                raise e
            raise

    def send_to_queue(self, queue: str, *args, **kwargs):
        durable = kwargs.pop("durable", True)
        with self.blocking_connection(connection_attempts=1) as conn:
            channel = conn.channel()
            channel.queue_declare(
                queue=queue,
                durable=durable,  # Persist across broker restarts.
            )
            return channel.send_to_queue(queue, *args, **kwargs)

    def tell_user(self, *args, **kwargs):
        with self.blocking_connection(connection_attempts=1) as conn:
            channel = conn.channel()
            return channel.tell_user(*args, **kwargs)

    # Streams

    def stream_connection_parameters(self, *, port: int = 5552, **kwargs):
        return self.queue_connection_parameters(port=port, **kwargs)

    def producer_connection(self, **kwargs) -> Producer:
        params = self.stream_connection_parameters(**kwargs)
        creds = params.credentials
        return Producer(
            host=params.host,
            username=creds.username,
            password=creds.password,
        )



import json
import logging
import zlib

from argparse import ArgumentParser
from dataclasses import dataclass
from socket import gethostname
from typing import Optional

from pika import BasicProperties
from pika.spec import Basic

from hive.messaging import blocking_connection

from ..channel import Channel

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(
        description="Churn messages from one queue to another.",
    )
    parser.add_argument(
        "source_queue", metavar="SOURCE",
        help="queue to consume from")
    parser.add_argument(
        "target_queue", metavar="TARGET", nargs="?",
        help="queue to publish to")
    parser.add_argument(
        "-v", "--verbose", action="count", default=0,
        help="increase verbosity (up to -vvv)")
    args = parser.parse_args()

    if (verbosity := args.verbose):
        logging.basicConfig(level={
            1: logging.WARNING,
            2: logging.INFO,
        }.get(verbosity, logging.DEBUG))

    worker = Worker(args.source_queue, args.target_queue)
    worker.run()


@dataclass
class Worker:
    source_queue: str
    target_queue: Optional[str] = None
    hostname: str = gethostname()
    mime_type: str = "application/vnd.gbenson.hive-messaging-smoke-test"

    @property
    def queues(self):
        yield self.source_queue
        if self.target_queue:
            yield self.target_queue

    def run(self):
        with blocking_connection() as conn:
            channel = conn.channel()
            self._run(channel)

    def _run(self, channel: Channel):
        for queue in self.queues:
            channel.queue_declare(
                queue=queue,
                durable=True,  # Persist across broker restarts.
            )
        channel.basic_qos(prefetch_count=1)  # Receive one message at a time.
        channel.basic_consume(
            queue=self.source_queue,
            on_message_callback=self._on_message,
        )
        channel.start_consuming()

    def _on_message(
            self,
            channel: Channel,
            method: Basic.Deliver,
            properties: BasicProperties,
            body: bytes,
    ):
        delivery_tag = method.delivery_tag
        try:
            content_type = properties.content_type
            if content_type == "application/json":
                exchange = method.exchange
                if exchange != "":
                    raise ValueError(exchange)

                routing_key = method.routing_key
                if routing_key != self.source_queue:
                    raise ValueError(routing_key)

                if not self.target_queue:
                    raise ValueError

                self._churn_forward(channel, body)

            elif content_type == self.mime_type:
                if self.target_queue:
                    raise ValueError(self.target_queue)

                self._churn_reversed(channel, body)

            else:
                raise ValueError(content_type)

            channel.basic_ack(delivery_tag=delivery_tag)

        except Exception:
            channel.basic_nack(delivery_tag=delivery_tag)
            logger.exception("EXCEPTION")
            raise  # XXX

    def _churn_forward(self, channel: Channel, body: bytes):
        msg = json.loads(body)

        if "routing_key" in msg:
            raise ValueError(msg["routing_key"])
        msg["routing_key"] = self.source_queue

        if "smoke_counts" not in msg:
            msg["smoke_counts"] = {}
        counts = msg["smoke_counts"]
        counts[self.hostname] = counts.get(self.hostname, 0) + 1

        body = zlib.compress(json.dumps(msg).encode("utf-8"))
        channel.send_to_queue(self.target_queue, body, self.mime_type)

    def _churn_reversed(self, channel: Channel, body: bytes):
        msg = json.loads(zlib.decompress(body))
        target_queue = msg.pop("routing_key")
        counts = msg["smoke_counts"]
        counts[self.hostname] = counts.get(self.hostname, 0) + 1
        channel.send_to_queue(target_queue, msg)



from __future__ import annotations

import weakref

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..channel import Channel


class ChannelService:
    def __init__(self, channel: Channel):
        self.channel = weakref.proxy(channel)

    def _declared_queue(self, queue, **kwargs) -> str:
        """Declare a queue, returning its name.

        Intended to be wrapped in a cached_property.
        """
        self.channel.queue_declare(queue, **kwargs)
        return queue



import json

from functools import cached_property
from typing import Callable, Optional

from pika import BasicProperties, DeliveryMode

from .wrapper import WrappedPikaThing
from .channel_services import Notifier


class Channel(WrappedPikaThing):
    @cached_property
    def notifier(self) -> Notifier:
        return Notifier(self)

    @cached_property
    def tell_user(self) -> Callable:
        return self.notifier.tell_user

    @cached_property
    def events_exchange(self) -> str:
        return self._hive_exchange(
            exchange="events",
            exchange_type="direct",
            durable=True,
        )

    @cached_property
    def dead_letter_exchange(self) -> str:
        return self._hive_exchange(
            exchange="dead.letter",
            exchange_type="direct",
            durable=True,
        )

    def _hive_exchange(self, exchange: str, **kwargs) -> str:
        name = f"hive.{exchange}"
        self.exchange_declare(exchange=name, **kwargs)
        return name

    def event_queue_declare(self, queue, **kwargs):
        result = self.queue_declare(queue, **kwargs)
        self.queue_bind(
            queue=queue,
            exchange=self.events_exchange,
            routing_key=queue,
        )
        return result

    def queue_declare(self, queue, **kwargs):
        dead_letter = kwargs.pop("dead_letter", False)
        if dead_letter:
            dead_letter_queue = f"x.{queue}"
            self.queue_declare(
                dead_letter_queue,
                durable=kwargs.get("durable", False),
            )

            dead_letter_exchange = self.dead_letter_exchange
            self.queue_bind(
                queue=dead_letter_queue,
                exchange=dead_letter_exchange,
                routing_key=queue,
            )

            arguments = kwargs.pop("arguments", {}).copy()
            self._ensure_arg(
                arguments,
                "x-dead-letter-exchange",
                dead_letter_exchange,
            )
            kwargs["arguments"] = arguments

        return self._pika.queue_declare(queue, **kwargs)

    def send_to_queue(
            self,
            queue: str,
            msg: bytes | dict,
            content_type: Optional[str] = None,
            **kwargs
    ):
        return self._publish(
            exchange="",
            routing_key=queue,
            message=msg,
            content_type=content_type,
            **kwargs
        )

    def publish_event(
            self,
            *,
            message: bytes | dict,
            routing_key: str,
            mandatory: bool,
            **kwargs
    ):
        return self._publish(
            message=message,
            exchange=self.events_exchange,
            routing_key=routing_key,
            mandatory=mandatory,
            **kwargs
        )

    def _publish(
            self,
            *,
            message: bytes | dict,
            exchange: str = "",
            routing_key: str = "",
            content_type: Optional[str] = None,
            delivery_mode: DeliveryMode = DeliveryMode.Persistent,
            mandatory: bool = True,
    ):
        payload, content_type = self._encapsulate(message, content_type)
        return self.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=payload,
            properties=BasicProperties(
                content_type=content_type,
                delivery_mode=delivery_mode,  # Persist across broker restarts.
            ),
            mandatory=mandatory,  # Don't fail silently.
        )

    @staticmethod
    def _ensure_arg(args: dict, key: str, want_value: any):
        if args.get(key, want_value) != want_value:
            raise ValueError(args)
        args[key] = want_value

    @staticmethod
    def _encapsulate(
            msg: bytes | dict,
            content_type: Optional[str],
    ) -> tuple[bytes, str]:
        """Prepare messages for transmission.
        """
        if not isinstance(msg, bytes):
            return json.dumps(msg).encode("utf-8"), "application/json"
        if not content_type:
            raise ValueError(f"content_type={content_type}")
        return msg, content_type

    def basic_consume(
            self,
            queue: str,
            on_message_callback,
            *args,
            **kwargs
    ):
        def _wrapped_callback(channel, *args, **kwargs):
            return on_message_callback(type(self)(channel), *args, **kwargs)
        return self._pika.basic_consume(
            queue=queue,
            on_message_callback=_wrapped_callback,
            *args,
            **kwargs
        )



import logging

from datetime import datetime

from hive import messaging as msgbus
from hive.common.testing import want_to_see


def test_send_to_queue(test_credentials, caplog):
    sent_msg = {"timestamp": str(datetime.now())}
    test_queue = "test.message_bus.send_recv"
    with caplog.at_level(logging.DEBUG, logger="pika"):
        msgbus.send_to_queue(test_queue, sent_msg)

    # Ensure we saw the broker confirm receipt of our message
    want_to_see(caplog, "Processing 1:Basic.Ack")

    # Ensure we initiated closing the channel
    want_to_see(caplog, "Closing connection (200): Normal shutdown")
